# Row Polymorphic Programming

```idris hide
module LessMacrosMoreTypes.RowPolymorphism

import Data.List
import Data.List.Elem
import Data.List.Quantifiers

import System

import DependentNuggets.FreshLists
```

Sometimes, especially when dealing with business logic, we have to deal with data that comes in from the real world, or elsewhere, naturally very messily typed, leading to nasty type signatures and messes of macros or code generation to build data structures from a schema. Row polymorphism can help us by abstracting over the fields contained in a record type in the type signature, letting us define records based on data, concatenate records together, and define functions that are generic across any record containing the required fields, all without macros.

This article is inspired by my work on implementing the _Brown Benchmark for Table Types_[^1] for my _Structures_[^2] library, an update for which will be released soon, providing a much more comprehensive collection of types for row-polymorphic programming.


## What is Row Polymorphism

At time of writing, the [Wikipedia page](https://en.wikipedia.org/wiki/Row_polymorphism) for row polymorphism starts with 
> In programming language type theory, row polymorphism is a kind of polymorphism that allows one to write programs that are polymorphic on row types such as record types and polymorphic variants.
and then leads into some mathematical notation heavy text that doesn't really elucidate much, and what the hell does "polymorphic on row types" even mean?

Consider a basic `record` in Idris as an example:
```idris
record ImACat where
  constructor MkCat
  name : String
  age : Nat
  meows_a_lot : Bool
```

This certainly seems like it might fit the bill, it says `record` right there. But what are these "row types"?

### Row Types

Row types often are called records in other contexts, and while "record" may be a somewhat more intuitive name for someone approaching this from the perspective of many common programming languages, I will continue to largely use the name "row type" as it will make more sense in the tabular data context I will be framing the rest of this post in terms of.

Consider some tabular data, say a table in a database, or a spread sheet. Tabular data is composed of a 2d grid of rows and columns, typically with each column having a name and potentially a type constraint on the values in that column. Consider this example of some tabular data: 

<style type="text/css">
#tbl-cats tbody tr :nth-child(2) {
    text-align: right;
}
</style>
<div class="md-table" id="tbl-cats">
<span class="md-table-title">Cats</span>

| name → String | age → Nat | meows_a_lot →  Bool |
|---------------|-----------|---------------------|
| Cricket       | 11        | False               |
| Cookie-Dough  | 4         | True                |
| Mango         | 0         | False               |
| Creampuff     | 5         | True                |

</div>

A row type is parameterized by a list of `(label, Type)` pairs, represented by the header of the table. We can see that the labels and types in our *Cats* table are identical to the fields of our `ImACat` record from earlier. Under this view, each row in the table can be thought of as an instance of the `ImACat` record. 

That parameterization on a schema, however, makes row types a bit more flexible, as well as a bit more complicated. We can perform the usual operations, like accessing a column in a row by name, but we can also perform new operations that enable a sort of limited, type-safe reflection over row types, such as selecting a specific set of columns from a row to produce a new row type, or concatenating the schemas of two rows to create a new row type expressing the result of joining the two rows.

### The Mysterious Row Polymorphism

Imagine you are helping your friend move and they need to figure out what furniture will fit in what rooms. The current property owner of their new place has graciously provided a table containing the length and width of each room, and their furniture suppler has provided a table detailing the lengths, widths, and heights of the bounding cuboids for each piece of furniture. Let's say the tables look a bit like this:

<style type="text/css">
#tbl-rooms tbody tr :nth-child(2),
#tbl-rooms tbody tr :nth-child(3) {
    text-align: right;
}
</style>
<div class="md-table" id="tbl-rooms">
<span class="md-table-title">Rooms</span>

| room → String | length → Double | width → Double |
|---------------|-----------------|----------------|
| Bedroom       | 3.65            | 3.35           |
| Living Room   | 4.87            | 3.35           |
| Dining Room   | 2.44            | 2.44           |
| Kitchen       | 3.35            | 2.44           |
| Bathroom      | 3.65            | 2.44           |

</div>

<style type="text/css">
#tbl-furniture tbody tr :nth-child(2), 
#tbl-furniture tbody tr :nth-child(3), 
#tbl-furniture tbody tr :nth-child(4) {
    text-align: right;
}
</style>
<div class="md-table" id="tbl-furniture">
<span class="md-table-title">Furniture</span>

| item → String | length → Double | width → Double | height → Double |
|---------------|-----------------|----------------|-----------------|
| Dresser       | 1.00            | 0.60           | 0.76            |
| Bed           | 2.00            | 1.50           | 0.50            |
| Nightstand    | 0.60            | 0.48           | 0.75            |
| Dinner Table  | 1.80            | 0.90           | 0.80            |
| TV Stand      | 1.93            | 0.42           | 0.90            |
| Coffee Table  | 1.81            | 0.81           | 0.46            |

</div>

One thing that's going to be critical to determine is how much floor space is left in a room after the furniture's been placed in it. After all, a room isn't of much use if there's not even enough free floor space to stand in it. You notice that both of these tables contain the information needed to calculate the available area and the foot prints of the furniture, the `length` and `width` columns.

Its tempting to define a `Room` data type and a `Furniture` data type containing each table's respective columns as fields, and describe the surface area/footprint via an `Area` interface. However, this approach has several drawbacks. It requires implementing the interface manually for each data type[^3], can hide certain categories of error like inadvertently using the wrong field to calculate the area, and is potentially quite fragile if the structure of the tables changes on us or we need to add more tables to our system.

We can do better. With our parameterized row types we can write an `area` function that is defined to work on any row type that has a `length` and a `width` column of type `Double`. This is the essence of row polymorphism, writing code that's generic not over what interfaces a type implements, but instead is generic over what fields the type contains.

### Pros and Cons

While row polymorphism does come with the downside of loosing proof-of-intentionality[^4] , so to speak, it does come with some major upsides. Two columns in two different tables having the same names and types, doesn't always mean that they are semantically equivalent, but its often a pretty good bet, especially when they are coming from the same source. More importantly, thinking about data types in terms of what values they contain rather than what behavior they implement is often a much better way to think about messy real-world data that was laid out with a dynamically typed language in mind on a good day, and not laid out with computers in mind at all on a bad day.

Like so many things in programming, its a trade off. Low level systems programming, where almost all the data you are working with was generated by the computer, often by systems you also control isn't going to get much benefit from row polymorphic programming. Other tasks like frontend, business logic, and video games quite frequently deal with extremely messy data that has only loosely defined shape and doesn't want to fit neatly into a type hierarchy have a lot to gain from row polymorphism. In the former category, you are more likely to hurt from the loss of semantic guarantees and end up needlessly weakening your type system and tripping over yourself more than you would without it. In the messier situation, you are instead likely to both end up saving type fighting with the type system trying to fit your data to it, and also run into the downsides of row polymorphism much less frequently.

Row polymorphism is also a fantastic way to think about assigning types to code that was originally written for a dynamically typed programming language[^5]. In my personal experience I have found that when an experienced programmer who really seems like they should know better is making some weird sounding arguments about how type systems decrease their productivity and aren't working it, they seem to usually be complaining about the lack of row polymorphism (or the closely related structural subtyping) in the popular statically typed languages, just without having the technical vocabulary for it.

## To Type a Table

One of the core features of row polymorphism is that the labels and types of the columns are determined by a piece of data, a list like structure containing pairs of labels and types. In languages built specifically to be row polymorphic, this data describing the column structure is often implicit. Idris, however, isn't built to be row polymorphic. It's type system just happens to be powerful enough that we can build it on our own, so we will have to build a schema composed of straight up pairs of labels and types.

A more developed library would probably use bespoke types for most of these components, but for the sake of keeping things concise, we'll use regular ol' lists as much as possible.

### Building the Schema

To make things just a little bit cleaner, we'll define our own data type for the pairs, which we will call `Header`:

> [!NOTE]
> Remember that the Idris concept called a `record` is a distinct concept, and Idris `record`s are _not_ the row types we are going to be building.

```idris
public export
record Header where
  constructor MkHeader
  label: String
  type: Type

export infixr 8 :>
public export
(:>) : String -> Type -> Header
(:>) str x = MkHeader str x
```

The schema itself is really simple, just an alias for a list of headers:

```idris
public export
Schema : Type
Schema = List Header
```


We will also need a way to compute the column types from our schema, turning a list of headers into a list of types:

> [!TIP]
> This function has an UpperCamelCase name because it is intended to be called in type signatures. This is both conventional, and reduces ambiguity as it can't be confused for a lowercase, implicitly defined, argument name.

```idris
public export
ColumnTypes : Schema -> List Type
ColumnTypes [] = []
ColumnTypes (h :: hs) = h.type :: ColumnTypes hs
```

### Making a row

A row is a list of the same length as the schema where every list element has a type determined by the header in the same position of the schema:

> [!TIP]
> `HList` is a standard library type with type constructor `List Type -> type`. Each type in the list of types determines the type of value in that position of the list. Notice this also constrains the length of the list, it must be exactly the same as the length of the list of types. This can be thought of as a more primitive version of the `Row` type we are building.

```idris
public export
Row : Schema -> Type
Row hs = HList (ColumnTypes hs)
```

We will also want a way of indexing the values in a row by their columns, instead of by their positions. We only want this function to be callable when the row actually _has_ that column, so we also require an inclusion proof, as an auto-implicit so the compiler can generate it for us when possible: 

> [!TIP]
> A value of type `Elem x xs` is a proof that the value `x` is included in the list `xs`, which also encodes its position in the list. The type signature of the `prf` argument thus reads "`h` is an element with known position in `hs`"

```idris
namespace RowIndex
  export
  index : {hs : Schema} -> (h : Header) -> (r : Row hs) -> {auto prf: h `Elem` hs} -> h.type
  index {hs = (_ :: hs)} h (x :: xs) {prf = Here} = x
  index {hs = (_ :: hs)} h (y :: ys) {prf = (There x)} = index h ys
```

### Making a table

The table itself is pretty simple, its just a list of rows:

```idris
public export
Table : Schema -> Type
Table hs = List (Row hs)
```

## Using our tables

Now that we have a way to describe the type signature of a table, let's go ahead and translate our previous table examples into code:

```idris
Rooms : Schema
Rooms = ["room" :> String, "length" :> Double, "width" :> Double]

rooms : Table Rooms
rooms = 
  [ ["Bedroom",     3.65, 3.35]
  , ["Living Room", 4.87, 3.35]
  , ["Dining Room", 2.44, 2.44]
  , ["Kitchen",     3.35, 2.44]
  , ["Bathroom",    3.65, 2.44]
  ]
```

```idris
Furniture : Schema
Furniture = ["item" :> String, "length" :> Double, "width" :> Double, "height" :> Double]

furniture : Table Furniture 
furniture = 
  [ ["Dresser",      1.00, 0.60, 0.76] 
  , ["Bed",          2.00, 1.50, 0.50]
  , ["Nightstand",   0.60, 0.48, 0.75]
  , ["Dinner Table", 1.80, 0.90, 0.80]
  , ["TV Stand",     1.93, 0.42, 0.90]
  , ["Coffee Table", 1.81, 0.81, 0.46]
  ]
```

### A simple row polymorphic function

With our tables defined, it's time to revisit our `area` function. The schema will be an implicit variable in the type signature, and we'll constrain it such that both the length and width fields are elements of the schema. These constraints allow us to simply call the `index` function with the given fields on the provided row:

> [!TIP]
> In a type signature, `x =>` is equivalent to `{auto _ : x} ->`

```idris
area : {hs : Schema} -> "length" :> Double `Elem` hs => "width" :> Double `Elem` hs => 
  Row hs -> Double
area x = 
  let length = ("length" :> Double) `index` x
      width = ("width" :> Double) `index` x
  in length * width
```

Without any further work, we can now call this function on both our room and furniture tables and it works as expected:

> [!NOTE]
> In this specific context, we do have to explicitly provide the schema to the `area` function to disambiguate, but this can be worked around and should be considered more of an artifact of the limited implementation in this blog post then an intrinsic ergonomics concern.

```idris hide
-- @@test map area rooms
mapAreaRooms : IO Bool
mapAreaRooms = do
  pure $
```

<div class="unit-test">
```idris
  map (area {hs = Rooms}) rooms == [12.2275, 16.314500000000002, 5.9536, 8.174, 8.905999999999999]
```
</div>

```idris hide
-- @@test map area furniture
mapAreaFurniture : IO Bool
mapAreaFurniture = do
  pure $
```

<div class="unit-test">
```idris
  map (area {hs = Furniture}) furniture == [0.6, 3.0, 0.288, 1.62, 0.8106, 1.4661000000000002]
```
</div>

### Building a new column

Just being able to access the fields in a row polymorphic way is pretty neat, but its also not really that impressive. We have much more interesting capabilities at our disposal, like adding new fields. We can take a function that consumes a row and produces a new value, along with a name for the new value, and use that to add a new column to each row in the table:

```idris
buildColumn : {hs : Schema} 
  -> (f : Row hs -> t) -> (name : String) -> Table hs -> Table (hs ++ [name :> t]) 
buildColumn f name table = map buildColumnRow table
  where
    addToEnd : {name' : String} -> {hs' : Schema} 
      -> t -> Row hs' -> Row (hs' ++ [name' :> t])
    addToEnd {hs' = []} x [] = [x]
    addToEnd {hs' = (y :: ys)} x (x' :: xs) = x' :: addToEnd x xs
    buildColumnRow : {name' : String} -> Row hs -> Row (hs ++ [name' :> t])
    buildColumnRow row = addToEnd (f row) row
```

With this new under our belt, we can make new tables derived from our previous ones, such as adding an area column to our previous two tables:

```idris
roomsWithArea : Table (Rooms ++ ["footprint" :> Double])
roomsWithArea = buildColumn {hs = Rooms} (area {hs = Rooms}) "footprint" rooms
```

```idris hide
-- @@test roomsWithArea
roomsWithAreaTest : IO Bool
roomsWithAreaTest = do
  pure $
```

<div class="unit-test">
```idris
  roomsWithArea ==
  [ ["Bedroom",     3.65, 3.35, 12.2275]
  , ["Living Room", 4.87, 3.35, 16.314500000000002]
  , ["Dining Room", 2.44, 2.44, 5.9536]
  , ["Kitchen",     3.35, 2.44, 8.174]
  , ["Bathroom",    3.65, 2.44, 8.905999999999999]
  ]
```
</div>

```idris
furnitureWithArea : Table (Furniture ++ ["footprint" :> Double])
furnitureWithArea = buildColumn {hs = Furniture} (area {hs = Furniture}) "footprint" furniture
```

```idris hide
-- @@test furnitureWithArea
furnitureWithAreaTest : IO Bool
furnitureWithAreaTest = do
  pure $
```

<div class="unit-test">
```idris
  furnitureWithArea ==
  [ ["Dresser",      1.0,  0.6,  0.76, 0.6]
  , ["Bed",          2.0,  1.5,  0.5, 3.0]
  , ["Nightstand",   0.6,  0.48, 0.75, 0.288]
  , ["Dinner Table", 1.8,  0.9,  0.8, 1.62]
  , ["TV Stand",     1.93, 0.42, 0.9, 0.8106]
  , ["Coffee Table", 1.81, 0.81, 0.46, 1.4661000000000002]
  ]
```
</div>

### Select

### Implementing Regular Interfaces for Row Types

### Generating table types for sql schemas, with runtime example

## Simplifications made for the sake of this blog post

- Hardcoding string as the label type
- Feilds aren't nullable
- Not validating the consistency of the schema

[^1]: <https://github.com/brownplt/B2T2>

[^2]: <https://git.sr.ht/~thatonelutenist/Structures>

[^3]: At best, we might be able to get away with something like a derive macro, but that's still something we have to invoke manually

[^4]: Generally speaking, with "traditional" data types, an interface typically requires some manual intervention from the programmer to be implemented for a given type. This provides a level of assurance that the implementation over that data type is intentional, and thus likely semantically valid.

[^5]: For instance, typescript's structural subtyping is a very good approximation of row polymorphism, and this is something I believe has contributued just as much to its success as a javascript replacement as its backwards compatibility. A statically typed langauge that could run existing javascript code, but made it a battle to assign types to that existing javascript code, wouldn't be very useful as a transition language.
